﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using System.IO;

using MPUSBAPI;


/*****************************************************************************************************************************
 * 
 *                  ISM 2.4 gHz Spectrum Analyzer
 *                  
 *     Version 2.0   August 2010 - C# Version - James E. McCullers
 *     
 *     Version 1.0   Nov 2009
 *
 *    Copyright 2009 Geoff Graham - http://geoffg.net
 *    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General
 *    Public License as published by the Free Software Foundation, either version 2 of the License, or (at your
 *    option) any later version.
 *
 *    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *    for more details.  You should have received a copy of the GNU General Public License along with this program.
 *    If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************************************************************/
namespace ISMScanner
{
    public partial class Form1 : Form
    {
        const int NbrChan = 85;                                 // Number of MHz (must be the same as in the 18F2550 firmware)
        const int ScaleMargin = 40;                             // Top/bottom margin for the freq scale 
        const int MaxSample = 750;

        byte[]  scannerData = new byte[NbrChan];
        byte[]  buff = new byte[NbrChan - 64];
        int[,]  da = new int[NbrChan, MaxSample];
        float   scaleBottom; 
        float   scaleWidth;
        bool    paused = false;
        bool    exporting = false;
        bool    drawScale = true;
        bool    closing = false;
        TextWriter textWriter;

        //  Functions and Data required from WIN32 API
        const long INVALID_HANDLE_VALUE = -1;
        const long ERROR_INVALID_HANDLE = 6;

        [System.Runtime.InteropServices.DllImport("kernel32")]
        private static extern UInt32 GetLastError();


        /*****************************************************************************************************************************
         *              Form
         *****************************************************************************************************************************/
        public Form1()
        {
            InitializeComponent();
            scaleBottom = pictureBox.Bottom - 3;
            scaleWidth = pictureBox.Width - 4;;

            exportButton.Click += OnExportButtonClick;
            pauseButton.Click += OnPauseButtonClick;
            holdTime.ValueChanged += OnHoldTimeValueChanged;
            pictureBox.Paint += OnPictureBoxPaint;
            this.Paint += OnPaint;
            this.FormClosing += OnClosing;
            this.Load += OnFormLoad;
        }

        /*****************************************************************************************************************************
         *              Form Load Event Handler
         *              
         *      Called when form opened and handles majority of work        
         *****************************************************************************************************************************/
        private void OnFormLoad(object sender, EventArgs e)
        {
            MPUSB       usb = new MPUSB();
            Graphics    g = pictureBox.CreateGraphics();
            long        result = 0;
 
            ClearDA();
            drawScale = true;
            this.Invalidate();
            sampleLabel.Text = String.Format(holdTime.Value + " readings");
            this.Show();

            usb.InPipe = "\\MCHP_EP1";
            usb.VidPid = "vid_04d8&pid_000b";               // Microchip USB Vendor - Device ID

            // Verify USB device DLL is in execution library
            try
            {
                result = usb.DLLVersion;
            }
            catch
            {
            if (result == 0)
                MessageBox.Show("Cannot find mpusbapi.dll.\nThis should be installed in the same directory as ISMScanner.exe",
                                            "ISMScanner",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error);               
            }

            // Attempt to open device.  Retry waiting on device to be inserted or installed
            while (!OpenScanner(usb))
            {
                Application.DoEvents();
            }

            // Main processing.  Read device data until error, device removed or form close button pressed
            while (GetLastError() != ERROR_INVALID_HANDLE && usb.DeviceCount > 0 && closing == false)
            {
                if (GetScannerData(usb))
                    if (!paused)
                        DisplayData(g);                    
                Application.DoEvents();
            }

            // Close USB connection and exit application
            usb.ClosePipe(MPUSB.PipeDir.Read);
            this.Close();
        }


        /*****************************************************************************************************************************
         *              Export Button Click Event Handler
         *****************************************************************************************************************************/
        private void OnExportButtonClick(object sender, EventArgs e)
        {
            var result = DialogResult.None;

            if (exporting)
            {
                exporting = false;
                exportButton.Text = "Export";
                textWriter.Close();                
            }

            else
            {
                result = MessageBox.Show("Data will be recorded in the executable directory as 'Spectrum.csv'",
                                             "ISMScanner",
                                             MessageBoxButtons.OKCancel,
                                             MessageBoxIcon.Question);

                if (result == DialogResult.OK)
                {
                    // Open file for output and write header line 
                    textWriter = new StreamWriter(@"Spectrum.csv");
                    textWriter.Write("2400 MHz,");

                    for (int i = 2401; i < 2400 + NbrChan - 1; i++)
                        textWriter.Write("{0} MHz,", i);
                    textWriter.WriteLine();

                    exporting = true;
                    exportButton.Text = "Stop Export";
                }
            }
        }


        /*****************************************************************************************************************************
         *              Form Close Button Click Event Handler
         *****************************************************************************************************************************/
        private void OnClosing(object sender, FormClosingEventArgs e)
        {
            closing = true;
        }


        /*****************************************************************************************************************************
         *             Pause Button Event Handler
         *****************************************************************************************************************************/
        private void OnPauseButtonClick(object sender, EventArgs e)
        {
            paused = !paused;
            if (paused)
            {
                pauseButton.Text = "Run";
                ClearDA();
            }
            else
                pauseButton.Text = "Pause";
        }


        /*****************************************************************************************************************************
         *             Hold time Slider Value Chage Event Handler
         *****************************************************************************************************************************/
         private void OnHoldTimeValueChanged(object sender, EventArgs e)
        {
            sampleLabel.Text = String.Format(holdTime.Value + " readings");
        }

        /*****************************************************************************************************************************
         *             Form Paint Event Handler
         ****************************************************************************************************************************/
        private void OnPaint(object sender, PaintEventArgs e)
        {
            // Repaint scale
            drawScale = true;       
        }


        /*****************************************************************************************************************************
         *             PictureBox Paint Event Handler
         *****************************************************************************************************************************/
        private void OnPictureBoxPaint(object sender, PaintEventArgs e)
        {
            if (drawScale)
            {
                DrawScale(e.Graphics);
                drawScale = false;
            }           
        }

        /*****************************************************************************************************************************
         *              Reset data array
         *****************************************************************************************************************************/
        private void ClearDA()
        {
            for (int i = 0; i < NbrChan; i++)
            {
                for (int j = 0; j < holdTime.Value; j++)
                {
                    if (i % 2 == 0)
                        da[i, j] = 0;
                    else
                        da[i, j] = 0;
                }
            }
        }


        /*****************************************************************************************************************************
         *              Draw the frequency scales on the display
         ****************************************************************************************************************************/
        private void DrawScale(Graphics g)
        {
            int     i;
            float   x;
            string  text;
            Font    font = new Font("Microsoft Sans Serif", 9, FontStyle.Regular);
            Pen     scalePen = new Pen(Brushes.Black, 1);
            PointF  pt1;
            PointF  pt2;

            text = "Frequency   (GHz)";
            SizeF size = g.MeasureString(text, font);
            x = (pictureBox.Width - (int)size.Width) / 2;
            g.DrawString(text, font, Brushes.Black, new PointF(x, 0));

            text = "802.11 b/g   Channel";
            size = g.MeasureString(text, font);
            x = (pictureBox.Width - (int)size.Width) / 2;
            g.DrawString(text, font, Brushes.Black, new PointF(x, scaleBottom - 49));  //temp on 52

            for (i = 0; i < NbrChan; i++)
            {
                x = i * (scaleWidth / NbrChan);

                // Draw the frequency scale
                if (i % 10 == 0)
                {
                    // Draw major tick
                    pt1 = new PointF(x + (scaleWidth / NbrChan) / 2, ScaleMargin - 7);
                    pt2 = new PointF(x + (scaleWidth / NbrChan) / 2, ScaleMargin - 2);
                    g.DrawLine(scalePen, pt1, pt2);

                    text = "2.4 " + String.Format("{0}", i / 10);
                    size = g.MeasureString(text, font);
                    float cx = x + ((scaleWidth / NbrChan) / 2) - (size.Width / 2);
                    if (i == 0)
                        cx = 0;

                    g.DrawString(text, font, Brushes.Black, new PointF(cx, 15));
                }
                else
                {
                    // Draw minor tick
                    pt1 = new PointF(x + (scaleWidth / NbrChan) / 2, ScaleMargin - 7);
                    pt2 = new PointF(x + (scaleWidth / NbrChan) / 2, ScaleMargin - 5);
                    g.DrawLine(scalePen, pt1, pt2);
                }

                // Draw the WiFi channel scale
                if (i >= 12 && i <= 72 && (i - 12) % 5 == 0)
                {
                    // Draw tick
                    pt1 = new PointF(x + (scaleWidth / NbrChan) / 2, scaleBottom - ScaleMargin - 30);
                    pt2 = new PointF(x + (scaleWidth / NbrChan) / 2, scaleBottom - ScaleMargin - 22);
                    g.DrawLine(scalePen, pt1, pt2);

                    text = String.Format("{0}", ((i - 12) / 5) + 1);
                    size = g.MeasureString(text, font);
                    float zx = x + (scaleWidth / NbrChan) / 2 - size.Width / 2;
                    g.DrawString(text, font, Brushes.Black, new PointF(zx, scaleBottom - ScaleMargin - 22));
                }
            }
        }


        /*****************************************************************************************************************************
         *              Displays data received from scanner
         ****************************************************************************************************************************/
        private void DisplayData(Graphics g)
        {
            int         i;
            int         j;
            int         p;
            float       x;
            float       y;
            PointF      pt1;
            PointF      pt2;
            SolidBrush  brush = new SolidBrush(Color.FromArgb(78, 137, 88));
            Pen         gPen = new Pen(brush, scaleWidth / NbrChan);
            Pen         wPen = new Pen(Brushes.White, scaleWidth / NbrChan);

            if (exporting)
                {
                    textWriter.Write("{0},", scannerData[0]);
                    for (i = 1; i < NbrChan - 1; i++)
                        textWriter.Write("{0},", scannerData[i]);
                    textWriter.WriteLine();
                }

                for (i = 0; i < NbrChan; i++)
                {
                    for (j = holdTime.Value - 1; j >= 0; j--)
                        da[i, j + 1] = da[i, j];

                    da[i, 0] = scannerData[i];

                    // Get the peak signal level
                    p = 0;
                    for (j = 0; j < holdTime.Value; j++)
                    {
                        if (da[i, j] > p)
                            p = da[i, j];
                    }

                    x = i * (scaleWidth / NbrChan);
                    if (p < 2)
                        p = 2;
                    y = (p - 2) * ((297 - (ScaleMargin * 2)) / 30) + ScaleMargin;

                    pt1 = new PointF(x, 297 - ScaleMargin);
                    pt2 = new PointF(x, 297 - y);
                    g.DrawLine(gPen, pt1, pt2);

                    pt1 = new PointF(x, 297 - y);
                    pt2 = new PointF(x, ScaleMargin);
                    g.DrawLine(wPen, pt1, pt2);
                }
              }


        /*****************************************************************************************************************************
         *              Open scanner for input
         ****************************************************************************************************************************/
        private bool OpenScanner(MPUSB usb)
        {
            usb.ClosePipe(MPUSB.PipeDir.Read);
            usb.OpenPipe(MPUSB.PipeDir.Read);

            if (usb.InStatus == INVALID_HANDLE_VALUE)
            {
                statusLabel.Text = "Scanner not found";
                return false;
            }
            else
            {
                statusLabel.Text = "Connected to Geoff's 2.4GHz Scanner";
                return true;
            }
        }


        /*****************************************************************************************************************************
         *             Read data from scanner
         *              
         *        Data is delivered in two packets, the first being 64 bytes and the second the remaining bytes
         *        of the message.  It is possible that the sender timed out on the first packet and we start by
         *        trying to read the second packet.  This is handled by the logic in this function.
         *    
         *       Return true when it has copied the message or false if there was an error
         ****************************************************************************************************************************/
        private bool GetScannerData(MPUSB   usb)
        {
        UInt32   expectedReceiveLength;
        UInt32   receiveLength;
    
        // State 1:  Try to read the first packet
        expectedReceiveLength = 64;
        if (usb.DeviceCount == 0)
            return false;

        if (usb.Read(scannerData, expectedReceiveLength, out receiveLength, 250) != MPUSB.SUCCESS)
            return false;
        if (receiveLength == 64)
            goto    STATE_3;
        if (receiveLength != NbrChan - 64)
            return false;
    
        // State 2:  Try again for the first packet
        expectedReceiveLength = 64;
        if (usb.DeviceCount == 0)
            return false;
       if (usb.Read(scannerData, expectedReceiveLength, out receiveLength, 250) != MPUSB.SUCCESS)
            return false;
        if (receiveLength != 64)
            return false;
            
STATE_3:
        // State 3:  Read the second packet
        expectedReceiveLength = NbrChan - 64;
        if (usb.DeviceCount == 0)
            return false;
        if (usb.Read(buff, expectedReceiveLength, out receiveLength, 250) != MPUSB.SUCCESS)
            return false;
        if (receiveLength != NbrChan - 64)
            return false;

        buff.CopyTo(scannerData, 64);    
        return true;
        }
    }
}


